@@ -3,7 +3,7 @@
    Entrypoint and utility functions
    Copyright (C) Matthew Chapman <matthewc.unsw.edu.au> 1999-2008
    Copyright 2002-2011 Peter Astrand <astrand@cendio.se> for Cendio AB
-   Copyright 2010-2013 Henrik Andersson <hean01@cendio.se> for Cendio AB
+   Copyright 2010-2014 Henrik Andersson <hean01@cendio.se> for Cendio AB
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -61,7 +61,8 @@ uint8 g_static_rdesktop_salt_16[16] = {
 
 char g_title[64] = "";
 char *g_username;
-char g_hostname[16];
+char g_password[64] = "";
+char g_hostname[16] = "";
 char g_keymapname[PATH_MAX] = "";
 unsigned int g_keylayout = 0x409;	/* Defaults to US keyboard layout */
 int g_keyboard_type = 0x4;	/* Defaults to US keyboard layout */
@@ -111,15 +112,21 @@ char g_seamless_spawn_cmd[512];
 RD_BOOL g_user_quit = False;
 uint32 g_embed_wnd;
 uint32 g_rdp5_performanceflags =
-	RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG | RDP5_NO_MENUANIMATIONS;
+	RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG | RDP5_NO_MENUANIMATIONS | RDP5_NO_CURSOR_SHADOW;
 /* Session Directory redirection */
 RD_BOOL g_redirect = False;
-char g_redirect_server[64];
-char g_redirect_domain[16];
-char g_redirect_password[64];
+char *g_redirect_server;
+uint32 g_redirect_server_len;
+char *g_redirect_domain;
+uint32 g_redirect_domain_len;
 char *g_redirect_username;
-char g_redirect_cookie[128];
+uint32 g_redirect_username_len;
+uint8 *g_redirect_lb_info;
+uint32 g_redirect_lb_info_len;
+uint8 *g_redirect_cookie;
+uint32 g_redirect_cookie_len;
 uint32 g_redirect_flags = 0;
+uint32 g_redirect_session_id = 0;
 
 uint32 g_reconnect_logonid = 0;
 char g_reconnect_random[16];
@@ -512,10 +519,9 @@ parse_server_and_port(char *server)
 int
 main(int argc, char *argv[])
 {
-	char server[64];
+	char server[256];
 	char fullhostname[64];
 	char domain[256];
-	char password[64];
 	char shell[256];
 	char directory[256];
 	RD_BOOL prompt_password, deactivated;
@@ -550,7 +556,7 @@ main(int argc, char *argv[])
 
 	flags = RDP_LOGON_NORMAL;
 	prompt_password = False;
-	g_seamless_spawn_cmd[0] = domain[0] = password[0] = shell[0] = directory[0] = 0;
+	g_seamless_spawn_cmd[0] = domain[0] = g_password[0] = shell[0] = directory[0] = 0;
 	g_embed_wnd = 0;
 
 	g_num_devices = 0;
@@ -617,7 +623,7 @@ main(int argc, char *argv[])
 					break;
 				}
 
-				STRNCPY(password, optarg, sizeof(password));
+				STRNCPY(g_password, optarg, sizeof(g_password));
 				flags |= RDP_LOGON_AUTO;
 
 				/* try to overwrite argument so it won't appear in ps */
@@ -1052,7 +1058,7 @@ main(int argc, char *argv[])
 		xfree(locale);
 
 
-	if (prompt_password && read_password(password, sizeof(password)))
+	if (prompt_password && read_password(g_password, sizeof(g_password)))
 		flags |= RDP_LOGON_AUTO;
 
 	if (g_title[0] == 0)
@@ -1062,7 +1068,7 @@ main(int argc, char *argv[])
 	}
 
 #ifdef RDP2VNC
-	rdp2vnc_connect(server, flags, domain, password, shell, directory);
+	rdp2vnc_connect(server, flags, domain, g_password, shell, directory);
 	return EX_OK;
 #else
 
@@ -1092,11 +1098,8 @@ main(int argc, char *argv[])
 		return EX_OSERR;
 
 #ifdef WITH_RDPSND
-	if (g_rdpsnd)
-	{
-		if (!rdpsnd_init(rdpsnd_optarg))
-			warning("Initializing sound-support failed!\n");
-	}
+	if (!rdpsnd_init(rdpsnd_optarg))
+		warning("Initializing sound-support failed!\n");
 #endif
 
 	if (g_lspci_enabled)
@@ -1114,13 +1117,21 @@ main(int argc, char *argv[])
 			xfree(g_username);
 			g_username = (char *) xmalloc(strlen(g_redirect_username) + 1);
 			STRNCPY(g_username, g_redirect_username, strlen(g_redirect_username) + 1);
-			STRNCPY(password, g_redirect_password, sizeof(password));
 			STRNCPY(server, g_redirect_server, sizeof(server));
 			flags |= RDP_LOGON_AUTO;
+
+			fprintf(stderr, "Redirected to %s@%s session %d.\n",
+				g_redirect_username, g_redirect_server, g_redirect_session_id);
+
+			/* A redirect on SSL from a 2003 WTS will result in a 'connection reset by peer'
+			   and therefor we just clear this error before we connect to redirected server.
+			 */
+			g_network_error = False;
 		}
 
 		ui_init_connection();
-		if (!rdp_connect(server, flags, domain, password, shell, directory, g_redirect))
+		if (!rdp_connect
+		    (server, flags, domain, g_password, shell, directory, g_reconnect_loop))
 		{
 
 			g_network_error = False;
@@ -1146,9 +1157,7 @@ main(int argc, char *argv[])
 		if (!g_packet_encryption)
 			g_encryption_initial = g_encryption = False;
 
-
 		DEBUG(("Connection successful.\n"));
-		memset(password, 0, sizeof(password));
 
 		rd_create_ui();
 		tcp_run_ui(True);
@@ -1664,6 +1673,7 @@ load_licence(unsigned char **data)
 	if (home == NULL)
 		return -1;
 
+	memset(hi, 0, sizeof(hi));
 	snprintf((char *) hi, 16, "%s", g_hostname);
 	sec_hash_sha1_16(ho, hi, g_static_rdesktop_salt_16);
 	sec_hash_to_string(hash, sizeof(hash), ho, sizeof(ho));
@@ -1712,6 +1722,7 @@ save_licence(unsigned char *data, int length)
 		return;
 	}
 
+	memset(hi, 0, sizeof(hi));
 	snprintf((char *) hi, 16, "%s", g_hostname);
 	sec_hash_sha1_16(ho, hi, g_static_rdesktop_salt_16);
 	sec_hash_to_string(hash, sizeof(hash), ho, sizeof(ho));
