@@ -1,10 +1,12 @@
 /* -*- c-basic-offset: 8 -*-
    rdesktop: A Remote Desktop Protocol client.
-   Copyright (C) Matthew Chapman 1999-2005
+   Copyright (C) Matthew Chapman <matthewc.unsw.edu.au> 1999-2008
+   Copyright 2004-2011 Peter Astrand <astrand@cendio.se> for Cendio AB
+   Copyright 2010-2013 Henrik Andersson <hean01@cendio.se> for Cendio AB
 
-   This program is free software; you can redistribute it and/or modify
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -13,8 +15,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 /*
@@ -60,13 +61,16 @@ extern DEVICE_FNS serial_fns;
 extern DEVICE_FNS printer_fns;
 extern DEVICE_FNS parallel_fns;
 extern DEVICE_FNS disk_fns;
+#ifdef WITH_SCARD
+extern DEVICE_FNS scard_fns;
+#endif
 extern FILEINFO g_fileinfo[];
-extern BOOL g_notify_stamp;
+extern RD_BOOL g_notify_stamp;
 
 static VCHANNEL *rdpdr_channel;
 
 /* If select() times out, the request for the device with handle g_min_timeout_fd is aborted */
-NTHANDLE g_min_timeout_fd;
+RD_NTHANDLE g_min_timeout_fd;
 uint32 g_num_devices;
 
 /* Table with information about rdpdr devices */
@@ -91,7 +95,7 @@ struct async_iorequest *g_iorequest;
 
 /* Return device_id for a given handle */
 int
-get_device_index(NTHANDLE handle)
+get_device_index(RD_NTHANDLE handle)
 {
 	int i;
 	for (i = 0; i < RDPDR_MAX_DEVICES; i++)
@@ -114,7 +118,7 @@ convert_to_unix_filename(char *filename)
 	}
 }
 
-static BOOL
+static RD_BOOL
 rdpdr_handle_ok(int device, int handle)
 {
 	switch (g_rdpdr_device[device].device_type)
@@ -135,7 +139,7 @@ rdpdr_handle_ok(int device, int handle)
 }
 
 /* Add a new io request to the table containing pending io requests so it won't block rdesktop */
-static BOOL
+static RD_BOOL
 add_async_iorequest(uint32 device, uint32 file, uint32 id, uint32 major, uint32 length,
 		    DEVICE_FNS * fns, uint32 total_timeout, uint32 interval_timeout, uint8 * buffer,
 		    uint32 offset)
@@ -309,13 +313,16 @@ rdpdr_send_available(void)
 	channel_send(s, rdpdr_channel);
 }
 
-static void
+void
 rdpdr_send_completion(uint32 device, uint32 id, uint32 status, uint32 result, uint8 * buffer,
 		      uint32 length)
 {
 	uint8 magic[4] = "rDCI";
 	STREAM s;
 
+#ifdef WITH_SCARD
+	scard_lock(SCARD_LOCK_RDPDR);
+#endif
 	s = channel_init(rdpdr_channel, 20 + length);
 	out_uint8a(s, magic, 4);
 	out_uint32_le(s, device);
@@ -330,6 +337,9 @@ rdpdr_send_completion(uint32 device, uint32 id, uint32 status, uint32 result, ui
 	/* hexdump(s->channel_hdr + 8, s->end - s->channel_hdr - 8); */
 #endif
 	channel_send(s, rdpdr_channel);
+#ifdef WITH_SCARD
+	scard_unlock(SCARD_LOCK_RDPDR);
+#endif
 }
 
 static void
@@ -352,12 +362,14 @@ rdpdr_process_irp(STREAM s)
 		error_mode,
 		share_mode, disposition, total_timeout, interval_timeout, flags_and_attributes = 0;
 
-	char filename[256];
+	char *filename;
+	uint32 filename_len;
+
 	uint8 *buffer, *pst_buf;
 	struct stream out;
 	DEVICE_FNS *fns;
-	BOOL rw_blocking = True;
-	NTSTATUS status = STATUS_INVALID_DEVICE_REQUEST;
+	RD_BOOL rw_blocking = True;
+	RD_NTSTATUS status = RD_STATUS_INVALID_DEVICE_REQUEST;
 
 	in_uint32_le(s, device);
 	in_uint32_le(s, file);
@@ -365,10 +377,20 @@ rdpdr_process_irp(STREAM s)
 	in_uint32_le(s, major);
 	in_uint32_le(s, minor);
 
+	filename = NULL;
+
 	buffer_len = 0;
 	buffer = (uint8 *) xmalloc(1024);
 	buffer[0] = 0;
 
+	if (device >= RDPDR_MAX_DEVICES)
+	{
+		error("invalid irp device 0x%lx file 0x%lx id 0x%lx major 0x%lx minor 0x%lx\n",
+		      device, file, id, major, minor);
+		xfree(buffer);
+		return;
+	}
+
 	switch (g_rdpdr_device[device].device_type)
 	{
 		case DEVICE_TYPE_SERIAL:
@@ -395,9 +417,15 @@ rdpdr_process_irp(STREAM s)
 			break;
 
 		case DEVICE_TYPE_SCARD:
+#ifdef WITH_SCARD
+			fns = &scard_fns;
+			rw_blocking = False;
+			break;
+#endif
 		default:
 
 			error("IRP for bad device %ld\n", device);
+			xfree(buffer);
 			return;
 	}
 
@@ -415,29 +443,29 @@ rdpdr_process_irp(STREAM s)
 
 			if (length && (length / 2) < 256)
 			{
-				rdp_in_unistr(s, filename, length);
-				convert_to_unix_filename(filename);
-			}
-			else
-			{
-				filename[0] = 0;
+				rdp_in_unistr(s, length, &filename, &filename_len);
+				if (filename)
+					convert_to_unix_filename(filename);
 			}
 
 			if (!fns->create)
 			{
-				status = STATUS_NOT_SUPPORTED;
+				status = RD_STATUS_NOT_SUPPORTED;
+				free(filename);
 				break;
 			}
 
 			status = fns->create(device, desired_access, share_mode, disposition,
 					     flags_and_attributes, filename, &result);
+
+			free(filename);
 			buffer_len = 1;
 			break;
 
 		case IRP_MJ_CLOSE:
 			if (!fns->close)
 			{
-				status = STATUS_NOT_SUPPORTED;
+				status = RD_STATUS_NOT_SUPPORTED;
 				break;
 			}
 
@@ -448,7 +476,7 @@ rdpdr_process_irp(STREAM s)
 
 			if (!fns->read)
 			{
-				status = STATUS_NOT_SUPPORTED;
+				status = RD_STATUS_NOT_SUPPORTED;
 				break;
 			}
 
@@ -459,7 +487,7 @@ rdpdr_process_irp(STREAM s)
 #endif
 			if (!rdpdr_handle_ok(device, file))
 			{
-				status = STATUS_INVALID_HANDLE;
+				status = RD_STATUS_INVALID_HANDLE;
 				break;
 			}
 
@@ -468,7 +496,7 @@ rdpdr_process_irp(STREAM s)
 				buffer = (uint8 *) xrealloc((void *) buffer, length);
 				if (!buffer)
 				{
-					status = STATUS_CANCELLED;
+					status = RD_STATUS_CANCELLED;
 					break;
 				}
 				status = fns->read(file, buffer, length, offset, &result);
@@ -480,7 +508,7 @@ rdpdr_process_irp(STREAM s)
 			pst_buf = (uint8 *) xmalloc(length);
 			if (!pst_buf)
 			{
-				status = STATUS_CANCELLED;
+				status = RD_STATUS_CANCELLED;
 				break;
 			}
 			serial_get_timeout(file, length, &total_timeout, &interval_timeout);
@@ -488,11 +516,11 @@ rdpdr_process_irp(STREAM s)
 			    (device, file, id, major, length, fns, total_timeout, interval_timeout,
 			     pst_buf, offset))
 			{
-				status = STATUS_PENDING;
+				status = RD_STATUS_PENDING;
 				break;
 			}
 
-			status = STATUS_CANCELLED;
+			status = RD_STATUS_CANCELLED;
 			break;
 		case IRP_MJ_WRITE:
 
@@ -500,7 +528,7 @@ rdpdr_process_irp(STREAM s)
 
 			if (!fns->write)
 			{
-				status = STATUS_NOT_SUPPORTED;
+				status = RD_STATUS_NOT_SUPPORTED;
 				break;
 			}
 
@@ -512,7 +540,7 @@ rdpdr_process_irp(STREAM s)
 #endif
 			if (!rdpdr_handle_ok(device, file))
 			{
-				status = STATUS_INVALID_HANDLE;
+				status = RD_STATUS_INVALID_HANDLE;
 				break;
 			}
 
@@ -526,7 +554,7 @@ rdpdr_process_irp(STREAM s)
 			pst_buf = (uint8 *) xmalloc(length);
 			if (!pst_buf)
 			{
-				status = STATUS_CANCELLED;
+				status = RD_STATUS_CANCELLED;
 				break;
 			}
 
@@ -535,18 +563,18 @@ rdpdr_process_irp(STREAM s)
 			if (add_async_iorequest
 			    (device, file, id, major, length, fns, 0, 0, pst_buf, offset))
 			{
-				status = STATUS_PENDING;
+				status = RD_STATUS_PENDING;
 				break;
 			}
 
-			status = STATUS_CANCELLED;
+			status = RD_STATUS_CANCELLED;
 			break;
 
 		case IRP_MJ_QUERY_INFORMATION:
 
 			if (g_rdpdr_device[device].device_type != DEVICE_TYPE_DISK)
 			{
-				status = STATUS_INVALID_HANDLE;
+				status = RD_STATUS_INVALID_HANDLE;
 				break;
 			}
 			in_uint32_le(s, info_level);
@@ -562,7 +590,7 @@ rdpdr_process_irp(STREAM s)
 
 			if (g_rdpdr_device[device].device_type != DEVICE_TYPE_DISK)
 			{
-				status = STATUS_INVALID_HANDLE;
+				status = RD_STATUS_INVALID_HANDLE;
 				break;
 			}
 
@@ -578,7 +606,7 @@ rdpdr_process_irp(STREAM s)
 
 			if (g_rdpdr_device[device].device_type != DEVICE_TYPE_DISK)
 			{
-				status = STATUS_INVALID_HANDLE;
+				status = RD_STATUS_INVALID_HANDLE;
 				break;
 			}
 
@@ -594,7 +622,7 @@ rdpdr_process_irp(STREAM s)
 
 			if (g_rdpdr_device[device].device_type != DEVICE_TYPE_DISK)
 			{
-				status = STATUS_INVALID_HANDLE;
+				status = RD_STATUS_INVALID_HANDLE;
 				break;
 			}
 
@@ -608,13 +636,11 @@ rdpdr_process_irp(STREAM s)
 					in_uint8s(s, 0x17);
 					if (length && length < 2 * 255)
 					{
-						rdp_in_unistr(s, filename, length);
-						convert_to_unix_filename(filename);
-					}
-					else
-					{
-						filename[0] = 0;
+						rdp_in_unistr(s, length, &filename, &filename_len);
+						if (filename)
+							convert_to_unix_filename(filename);
 					}
+
 					out.data = out.p = buffer;
 					out.size = sizeof(buffer);
 					status = disk_query_directory(file, info_level, filename,
@@ -622,6 +648,8 @@ rdpdr_process_irp(STREAM s)
 					result = buffer_len = out.p - out.data;
 					if (!buffer_len)
 						buffer_len++;
+
+					free(filename);
 					break;
 
 				case IRP_MN_NOTIFY_CHANGE_DIRECTORY:
@@ -631,19 +659,17 @@ rdpdr_process_irp(STREAM s)
 
 					in_uint32_le(s, info_level);	/* notify mask */
 
-					g_notify_stamp = True;
-
 					status = disk_create_notify(file, info_level);
 					result = 0;
 
-					if (status == STATUS_PENDING)
+					if (status == RD_STATUS_PENDING)
 						add_async_iorequest(device, file, id, major, length,
 								    fns, 0, 0, NULL, 0);
 					break;
 
 				default:
 
-					status = STATUS_INVALID_PARAMETER;
+					status = RD_STATUS_INVALID_PARAMETER;
 					/* JIF */
 					unimpl("IRP major=0x%x minor=0x%x\n", major, minor);
 			}
@@ -653,7 +679,7 @@ rdpdr_process_irp(STREAM s)
 
 			if (!fns->device_control)
 			{
-				status = STATUS_NOT_SUPPORTED;
+				status = RD_STATUS_NOT_SUPPORTED;
 				break;
 			}
 
@@ -665,25 +691,33 @@ rdpdr_process_irp(STREAM s)
 			buffer = (uint8 *) xrealloc((void *) buffer, bytes_out + 0x14);
 			if (!buffer)
 			{
-				status = STATUS_CANCELLED;
+				status = RD_STATUS_CANCELLED;
 				break;
 			}
 
 			out.data = out.p = buffer;
 			out.size = sizeof(buffer);
+
+#ifdef WITH_SCARD
+			scardSetInfo(device, id, bytes_out + 0x14);
+#endif
 			status = fns->device_control(file, request, s, &out);
 			result = buffer_len = out.p - out.data;
 
 			/* Serial SERIAL_WAIT_ON_MASK */
-			if (status == STATUS_PENDING)
+			if (status == RD_STATUS_PENDING)
 			{
 				if (add_async_iorequest
 				    (device, file, id, major, length, fns, 0, 0, NULL, 0))
 				{
-					status = STATUS_PENDING;
+					status = RD_STATUS_PENDING;
 					break;
 				}
 			}
+#ifdef WITH_SCARD
+			else if (status == (RD_STATUS_PENDING | 0xC0000000))
+				status = RD_STATUS_PENDING;
+#endif
 			break;
 
 
@@ -691,7 +725,7 @@ rdpdr_process_irp(STREAM s)
 
 			if (g_rdpdr_device[device].device_type != DEVICE_TYPE_DISK)
 			{
-				status = STATUS_INVALID_HANDLE;
+				status = RD_STATUS_INVALID_HANDLE;
 				break;
 			}
 
@@ -701,7 +735,7 @@ rdpdr_process_irp(STREAM s)
 			out.size = sizeof(buffer);
 			/* FIXME: Perhaps consider actually *do*
 			   something here :-) */
-			status = STATUS_SUCCESS;
+			status = RD_STATUS_SUCCESS;
 			result = buffer_len = out.p - out.data;
 			break;
 
@@ -710,7 +744,7 @@ rdpdr_process_irp(STREAM s)
 			break;
 	}
 
-	if (status != STATUS_PENDING)
+	if (status != RD_STATUS_PENDING)
 	{
 		rdpdr_send_completion(device, id, status, result, buffer, buffer_len);
 	}
@@ -720,7 +754,7 @@ rdpdr_process_irp(STREAM s)
 }
 
 static void
-rdpdr_send_clientcapabilty(void)
+rdpdr_send_clientcapability(void)
 {
 	uint8 magic[4] = "rDPC";
 	STREAM s;
@@ -787,7 +821,7 @@ rdpdr_process(STREAM s)
 		if ((magic[2] == 'C') && (magic[3] == 'C'))
 		{
 			/* connect from server */
-			rdpdr_send_clientcapabilty();
+			rdpdr_send_clientcapability();
 			rdpdr_send_available();
 			return;
 		}
@@ -817,23 +851,20 @@ rdpdr_process(STREAM s)
 	unimpl("RDPDR packet type %c%c%c%c\n", magic[0], magic[1], magic[2], magic[3]);
 }
 
-BOOL
+RD_BOOL
 rdpdr_init()
 {
-	if (g_num_devices > 0)
-	{
-		rdpdr_channel =
-			channel_register("rdpdr",
-					 CHANNEL_OPTION_INITIALIZED | CHANNEL_OPTION_COMPRESS_RDP,
-					 rdpdr_process);
-	}
+	rdpdr_channel =
+		channel_register("rdpdr",
+				 CHANNEL_OPTION_INITIALIZED | CHANNEL_OPTION_COMPRESS_RDP,
+				 rdpdr_process);
 
 	return (rdpdr_channel != NULL);
 }
 
 /* Add file descriptors of pending io request to select() */
 void
-rdpdr_add_fds(int *n, fd_set * rfds, fd_set * wfds, struct timeval *tv, BOOL * timeout)
+rdpdr_add_fds(int *n, fd_set * rfds, fd_set * wfds, struct timeval *tv, RD_BOOL * timeout)
 {
 	uint32 select_timeout = 0;	/* Timeout value to be used for select() (in millisecons). */
 	struct async_iorequest *iorq;
@@ -929,10 +960,10 @@ rdpdr_remove_iorequest(struct async_iorequest *prev, struct async_iorequest *ior
 }
 
 /* Check if select() returned with one of the rdpdr file descriptors, and complete io if it did */
-void
-_rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out)
+static void
+_rdpdr_check_fds(fd_set * rfds, fd_set * wfds, RD_BOOL timed_out)
 {
-	NTSTATUS status;
+	RD_NTSTATUS status;
 	uint32 result = 0;
 	DEVICE_FNS *fns;
 	struct async_iorequest *iorq;
@@ -961,7 +992,7 @@ _rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out)
 					/* iv_timeout between 2 chars, send partial_len */
 					/*printf("RDPDR: IVT total %u bytes read of %u\n", iorq->partial_len, iorq->length); */
 					rdpdr_send_completion(iorq->device,
-							      iorq->id, STATUS_SUCCESS,
+							      iorq->id, RD_STATUS_SUCCESS,
 							      iorq->partial_len,
 							      iorq->buffer, iorq->partial_len);
 					iorq = rdpdr_remove_iorequest(prev, iorq);
@@ -985,7 +1016,7 @@ _rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out)
 
 		}
 
-		rdpdr_abort_io(g_min_timeout_fd, 0, STATUS_TIMEOUT);
+		rdpdr_abort_io(g_min_timeout_fd, 0, RD_STATUS_TIMEOUT);
 		return;
 	}
 
@@ -1089,7 +1120,7 @@ _rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out)
 						out.size = sizeof(buffer);
 						out_uint32_le(&out, result);
 						result = buffer_len = out.p - out.data;
-						status = STATUS_SUCCESS;
+						status = RD_STATUS_SUCCESS;
 						rdpdr_send_completion(iorq->device, iorq->id,
 								      status, result, buffer,
 								      buffer_len);
@@ -1125,7 +1156,7 @@ _rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out)
 						{
 							g_notify_stamp = False;
 							status = disk_check_notify(iorq->fd);
-							if (status != STATUS_PENDING)
+							if (status != RD_STATUS_PENDING)
 							{
 								rdpdr_send_completion(iorq->device,
 										      iorq->id,
@@ -1151,7 +1182,7 @@ _rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out)
 }
 
 void
-rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out)
+rdpdr_check_fds(fd_set * rfds, fd_set * wfds, RD_BOOL timed_out)
 {
 	fd_set dummy;
 
@@ -1169,8 +1200,8 @@ rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out)
 
 
 /* Abort a pending io request for a given handle and major */
-BOOL
-rdpdr_abort_io(uint32 fd, uint32 major, NTSTATUS status)
+RD_BOOL
+rdpdr_abort_io(uint32 fd, uint32 major, RD_NTSTATUS status)
 {
 	uint32 result;
 	struct async_iorequest *iorq;
